@@ -6,26 +6,29 @@ use std::{collections::HashMap, fs, io};
 
 use anyhow::{bail, Context, Result};
 use async_trait::async_trait;
-use futures::join;
 use futures::StreamExt;
+use futures::{join, Future};
+use jetty_core::cual::Cual;
 use jetty_core::logging::error;
 use serde::{Deserialize, Serialize};
+use serde_with::serde_as;
 
 use crate::file_parse::origin::SourceOrigin;
-use crate::nodes::{self, Permissionable, ProjectId};
+use crate::nodes::{self, Permissionable, ProjectId, TableauCualable};
 
-use crate::rest;
+use crate::rest::{self, TableauAssetType};
 use crate::TableauCredentials;
 
 /// Number of assets to download concurrently
 const CONCURRENT_ASSET_DOWNLOADS: usize = 25;
 /// Number of metadata request to run currently (e.g. permissions)
-const CONCURRENT_METADATA_FETCHES: usize = 100;
+pub(crate) const CONCURRENT_METADATA_FETCHES: usize = 100;
 /// Path to serialized version of the Tableau Env
 const SERIALIZED_ENV_FILENAME: &str = "tableau_env.json";
 
 /// The state of a tableau site. We use this to persist state and
 /// enable incremental updates.
+#[serde_as]
 #[derive(Default, Deserialize, Serialize, Debug, Clone)]
 pub(crate) struct Environment {
     pub users: HashMap<String, nodes::User>,
@@ -37,6 +40,14 @@ pub(crate) struct Environment {
     pub metrics: HashMap<String, nodes::Metric>,
     pub views: HashMap<String, nodes::View>,
     pub workbooks: HashMap<String, nodes::Workbook>,
+    #[serde_as(as = "HashMap<serde_with::json::JsonString, _>")]
+    pub cual_id_map: HashMap<Cual, TableauAssetReference>,
+}
+
+#[derive(Hash, Default, Deserialize, Serialize, Debug, Clone, PartialEq, Eq)]
+pub(crate) struct TableauAssetReference {
+    pub(crate) asset_type: TableauAssetType,
+    pub(crate) id: String,
 }
 
 impl Environment {
@@ -52,6 +63,17 @@ impl Environment {
         }
         res
     }
+
+    /// given a group name, return the group id
+    pub(crate) fn get_group_id_by_name(&self, group_name: &String) -> Option<String> {
+        self.groups.iter().find_map(|(id, g)| {
+            if &g.name == group_name {
+                Some(id.to_owned())
+            } else {
+                None
+            }
+        })
+    }
 }
 
 /// Implemented for asset types that have sources embedded in them: Workbooks, Flows, and Datasources
@@ -202,21 +224,25 @@ impl Coordinator {
                 error!("unable to fetch groups: {}", e);
                 Default::default()
             }),
+            // FUTURE: update all calls to create a cual to just use this. Probably easier to have it centralized
+            cual_id_map: Default::default(),
         };
 
-        // Now, make sure that assets sources are all up to date
+        // FUTURE: put this back in. Right now, we don't fetch the sources well, so we'll just skip it. This needs to be
+        // reworked to use the metadata api
+        // // Now, make sure that assets sources are all up to date
 
-        let source_futures = vec![
-            self.get_source_futures_from_map(&mut new_env.flows, &self.env.flows),
-            self.get_source_futures_from_map(&mut new_env.datasources, &self.env.datasources),
-            self.get_source_futures_from_map(&mut new_env.workbooks, &self.env.workbooks),
-        ];
+        // let source_futures = vec![
+        //     self.get_source_futures_from_map(&mut new_env.flows, &self.env.flows),
+        //     self.get_source_futures_from_map(&mut new_env.datasources, &self.env.datasources),
+        //     self.get_source_futures_from_map(&mut new_env.workbooks, &self.env.workbooks),
+        // ];
 
-        // Source fetches
-        futures::stream::iter(source_futures.into_iter().flatten())
-            .buffer_unordered(CONCURRENT_ASSET_DOWNLOADS)
-            .collect::<Vec<_>>()
-            .await;
+        // // Source fetches
+        // futures::stream::iter(source_futures.into_iter().flatten())
+        //     .buffer_unordered(CONCURRENT_ASSET_DOWNLOADS)
+        //     .collect::<Vec<_>>()
+        //     .await;
 
         // Clone the env so we don't try to both immutably and mutably borrow at the same time.
         let new_env_clone = new_env.clone();
@@ -232,15 +258,45 @@ impl Coordinator {
         ];
 
         // Permission fetches
-        futures::stream::iter(permission_futures.into_iter().flatten())
+        let fetch_results = futures::stream::iter(permission_futures.into_iter().flatten())
             .buffer_unordered(CONCURRENT_METADATA_FETCHES)
             .collect::<Vec<_>>()
             .await;
 
+        fetch_results.into_iter().for_each(|r| match r {
+            Ok(_) => (),
+            Err(e) => error!("problem fetching tableau permissions: {e}"),
+        });
+
+        // Default permission fetches
+        let fetch_results = futures::stream::iter(
+            self.get_default_permission_futures_for_projects(&mut new_env.projects, &new_env_clone),
+        )
+        .buffer_unordered(CONCURRENT_METADATA_FETCHES)
+        .collect::<Vec<_>>()
+        .await;
+
+        fetch_results.into_iter().for_each(|r| match r {
+            Ok(_) => (),
+            Err(e) => error!("problem fetching tableau default permissions: {e}"),
+        });
+
         // get group membership
-        self.get_groups_users(&mut new_env.groups, &new_env.users)
+        let group_membership_futures =
+            self.get_groups_users(&mut new_env.groups, &new_env_clone.users);
+
+        let fetch_results = futures::stream::iter(group_membership_futures)
+            .buffer_unordered(CONCURRENT_METADATA_FETCHES)
+            .collect::<Vec<_>>()
             .await;
 
+        fetch_results.into_iter().for_each(|r| match r {
+            Ok(_) => (),
+            Err(e) => error!("problem fetching tableau groups: {e}"),
+        });
+
+        // update the cual map
+        build_cual_map(&mut new_env);
         // update self.env
         self.env = new_env;
 
@@ -257,23 +313,21 @@ impl Coordinator {
 
     /// Get all the users for a map of groups. Returns a Vec of results that
     /// can be checked to know if any group membership was not fetched successfully
-    async fn get_groups_users(
-        &self,
-        groups: &mut HashMap<String, nodes::Group>,
-        users: &HashMap<String, nodes::User>,
-    ) -> Vec<Result<(), anyhow::Error>> {
-        futures::stream::iter(
-            groups
-                .iter_mut()
-                .map(|(_, v)| v.update_users(&self.rest_client, users)),
-        )
-        .buffer_unordered(CONCURRENT_METADATA_FETCHES)
-        .collect::<Vec<_>>()
-        .await
+    fn get_groups_users<'b>(
+        &'b self,
+        groups: &'b mut HashMap<String, nodes::Group>,
+        users: &'b HashMap<String, nodes::User>,
+    ) -> Vec<impl futures::Future<Output = Result<()>> + 'b> {
+        let fetches = groups
+            .values_mut()
+            .map(|d| d.update_users(&self.rest_client, users))
+            .collect::<Vec<_>>();
+        fetches
     }
 
     /// Return a Vec of futures (sort of - look at return type) that will fetch futures from
     /// a map of assets
+    #[allow(clippy::type_complexity)]
     fn get_source_futures_from_map<'a, T: HasSources + Send + Sync>(
         &'a self,
         new_assets: &'a mut HashMap<String, T>,
@@ -295,6 +349,7 @@ impl Coordinator {
     }
 
     /// Return a Vec of futures that will request permissions for a collection of assets
+    #[allow(clippy::type_complexity)]
     fn get_permission_futures_from_map<'a, T: Permissionable + Send>(
         &'a self,
         new_assets: &'a mut HashMap<String, T>,
@@ -314,6 +369,19 @@ impl Coordinator {
             .collect::<Vec<_>>();
         fetches
     }
+
+    /// Return a Vec of futures that will request permissions for a collection of assets
+    fn get_default_permission_futures_for_projects<'a>(
+        &'a self,
+        projects: &'a mut HashMap<String, nodes::Project>,
+        env: &'a Environment,
+    ) -> Vec<impl Future<Output = Result<(), anyhow::Error>> + 'a> {
+        let fetches = projects
+            .values_mut()
+            .map(|v| v.update_default_permissions(&self.rest_client, env))
+            .collect::<Vec<_>>();
+        fetches
+    }
 }
 
 /// Read and parse the saved Tableau environment asset information
@@ -329,6 +397,72 @@ fn read_environment_assets(data_dir: PathBuf) -> Result<Environment> {
     Ok(e)
 }
 
+fn build_cual_map(env: &mut Environment) {
+    for (id, node) in &env.projects {
+        env.cual_id_map.insert(
+            node.cual(env),
+            TableauAssetReference {
+                asset_type: TableauAssetType::Project,
+                id: id.to_owned(),
+            },
+        );
+    }
+    for (id, node) in &env.datasources {
+        env.cual_id_map.insert(
+            node.cual(env),
+            TableauAssetReference {
+                asset_type: TableauAssetType::Datasource,
+                id: id.to_owned(),
+            },
+        );
+    }
+    for (id, node) in &env.flows {
+        env.cual_id_map.insert(
+            node.cual(env),
+            TableauAssetReference {
+                asset_type: TableauAssetType::Flow,
+                id: id.to_owned(),
+            },
+        );
+    }
+    for (id, node) in &env.lenses {
+        env.cual_id_map.insert(
+            node.cual(env),
+            TableauAssetReference {
+                asset_type: TableauAssetType::Lens,
+                id: id.to_owned(),
+            },
+        );
+    }
+    for (id, node) in &env.metrics {
+        env.cual_id_map.insert(
+            node.cual(env),
+            TableauAssetReference {
+                asset_type: TableauAssetType::Metric,
+                id: id.to_owned(),
+            },
+        );
+    }
+    for (id, node) in &env.views {
+        env.cual_id_map.insert(
+            node.cual(env),
+            TableauAssetReference {
+                asset_type: TableauAssetType::View,
+                id: id.to_owned(),
+            },
+        );
+    }
+    for (id, node) in &env.workbooks {
+        env.cual_id_map.insert(
+            node.cual(env),
+            TableauAssetReference {
+                asset_type: TableauAssetType::Workbook,
+                id: id.to_owned(),
+            },
+        );
+    }
+}
+
 #[cfg(test)]
 mod test {
     use crate::nodes::Project;
@@ -348,6 +482,8 @@ mod test {
                     Some(ProjectId("project1".to_owned())),
                     None,
                     vec![],
+                    Default::default(),
+                    Default::default(),
                 ),
             ),
             (
@@ -359,6 +495,8 @@ mod test {
                     Some(ProjectId("project2".to_owned())),
                     None,
                     vec![],
+                    Default::default(),
+                    Default::default(),
                 ),
             ),
             (
@@ -370,6 +508,8 @@ mod test {
                     None,
                     None,
                     vec![],
+                    Default::default(),
+                    Default::default(),
                 ),
             ),
         ]);
