@@ -2,41 +2,51 @@
 //!
 
 #![deny(missing_docs)]
-#![allow(dead_code)]
 
 mod coordinator;
 mod file_parse;
 mod nodes;
 mod permissions;
-mod rest;
+pub(crate) mod rest;
+mod write;
 
 use anyhow::{anyhow, Context, Result};
 use async_trait::async_trait;
 
 use coordinator::Environment;
+use futures::StreamExt;
 use rest::get_cual_prefix;
 pub use rest::TableauRestClient;
 use serde::{Deserialize, Serialize};
 use serde_json::json;
 
 use jetty_core::{
+    access_graph::translate::diffs::LocalConnectorDiffs,
     connectors::{
         nodes::ConnectorData,
         nodes::{self as jetty_nodes, EffectivePermission, SparseMatrix},
-        ConnectorClient, NewConnector,
+        AssetType, ConnectorCapabilities, ConnectorClient, NewConnector, ReadCapabilities,
+        WriteCapabilities,
     },
     cual::Cual,
-    jetty::{ConnectorConfig, CredentialsMap},
+    jetty::{ConnectorConfig, ConnectorManifest, CredentialsMap},
+    logging::error,
     permissions::matrix::Merge,
     Connector,
 };
 
 use nodes::{asset_to_policy::env_to_jetty_policies, FromTableau};
-use permissions::PermissionManager;
+use permissions::{
+    consts::{
+        DATASOURCE_CAPABILITIES, FLOW_CAPABILITIES, LENS_CAPABILITIES, METRIC_CAPABILITIES,
+        PROJECT_CAPABILITIES, VIEW_CAPABILITIES, WORKBOOK_CAPABILITIES,
+    },
+    PermissionManager,
+};
 
 use std::{
     collections::{HashMap, HashSet},
-    path::{PathBuf},
+    path::PathBuf,
 };
 
 /// Map wrapper for config values.
@@ -51,8 +61,8 @@ pub struct TableauCredentials {
     #[serde(flatten)]
     method: LoginMethod,
     /// Tableau server name like 10ay.online.tableau.com *without* the `https://`
-    server_name: String,
-    site_name: String,
+    pub(crate) server_name: String,
+    pub(crate) site_name: String,
 }
 
 /// Enum representing the different types of tableau login methods and the required information
@@ -110,10 +120,9 @@ impl TableauCredentials {
 }
 
 /// Top-level connector struct.
-#[allow(dead_code)]
 #[derive(Default)]
 pub struct TableauConnector {
-    config: TableauConfig,
+    _config: TableauConfig,
     coordinator: coordinator::Coordinator,
 }
 
@@ -125,6 +134,7 @@ impl TableauConnector {
     }
 
     /// Get the environment, but transformed into Jetty objects.
+    #[allow(clippy::type_complexity)]
     fn env_to_jetty_all(
         &self,
     ) -> (
@@ -133,6 +143,7 @@ impl TableauConnector {
         Vec<jetty_nodes::RawAsset>,
         Vec<jetty_nodes::RawTag>,
         Vec<jetty_nodes::RawPolicy>,
+        Vec<jetty_nodes::RawDefaultPolicy>,
     ) {
         // Transform assets
         let flows: Vec<jetty_nodes::RawAsset> =
@@ -159,12 +170,22 @@ impl TableauConnector {
         // Transform policies
         let all_policies = env_to_jetty_policies(&self.coordinator.env);
 
+        // Get default policies for each project
+        let default_policies = self
+            .coordinator
+            .env
+            .projects
+            .iter()
+            .flat_map(|(_, project)| project.get_default_policies(&self.coordinator.env))
+            .collect();
+
         (
             self.to_jetty(&self.coordinator.env.groups),
             self.to_jetty(&self.coordinator.env.users),
             all_assets,
             vec![], // self.object_to_jetty(&self.coordinator.env.tags);
             all_policies,
+            default_policies,
         )
     }
 
@@ -235,7 +256,7 @@ impl NewConnector for TableauConnector {
         let creds = TableauCredentials::from_map(credentials)?;
 
         let tableau_connector = TableauConnector {
-            config: config.config.to_owned(),
+            _config: config.config.to_owned(),
             coordinator: coordinator::Coordinator::new(creds, data_dir).await?,
         };
 
@@ -250,22 +271,139 @@ impl Connector for TableauConnector {
     }
 
     async fn get_data(&mut self) -> ConnectorData {
-        let (groups, users, assets, tags, policies) = self.env_to_jetty_all();
+        self.setup().await.unwrap();
+        let (groups, users, assets, tags, policies, default_policies) = self.env_to_jetty_all();
         let effective_permissions = self.get_effective_permissions();
-        ConnectorData::new(
+        ConnectorData {
             groups,
             users,
             assets,
             tags,
             policies,
-            Default::default(),
+            default_policies,
+            asset_references: Default::default(),
             effective_permissions,
-            Some(
+            cual_prefix: Some(
                 get_cual_prefix()
                     .context("tableau cual prefix not yet set")
                     .unwrap()
                     .to_owned(),
             ),
-        )
+        }
+    }
+
+    fn get_manifest(&self) -> ConnectorManifest {
+        let asset_privileges = [
+            (
+                AssetType("workbook".to_owned()),
+                WORKBOOK_CAPABILITIES
+                    .iter()
+                    .flat_map(|v| [format!("Allow{v}"), format!("Deny{v}")])
+                    .collect(),
+            ),
+            (
+                AssetType("lens".to_owned()),
+                LENS_CAPABILITIES
+                    .iter()
+                    .flat_map(|v| [format!("Allow{v}"), format!("Deny{v}")])
+                    .collect(),
+            ),
+            (
+                AssetType("datasource".to_owned()),
+                DATASOURCE_CAPABILITIES
+                    .iter()
+                    .flat_map(|v| [format!("Allow{v}"), format!("Deny{v}")])
+                    .collect(),
+            ),
+            (
+                AssetType("flow".to_owned()),
+                FLOW_CAPABILITIES
+                    .iter()
+                    .flat_map(|v| [format!("Allow{v}"), format!("Deny{v}")])
+                    .collect(),
+            ),
+            (
+                AssetType("metric".to_owned()),
+                METRIC_CAPABILITIES
+                    .iter()
+                    .flat_map(|v| [format!("Allow{v}"), format!("Deny{v}")])
+                    .collect(),
+            ),
+            (
+                AssetType("project".to_owned()),
+                PROJECT_CAPABILITIES
+                    .iter()
+                    .flat_map(|v| [format!("Allow{v}"), format!("Deny{v}")])
+                    .collect(),
+            ),
+            (
+                AssetType("view".to_owned()),
+                VIEW_CAPABILITIES
+                    .iter()
+                    .flat_map(|v| [format!("Allow{v}"), format!("Deny{v}")])
+                    .collect(),
+            ),
+        ]
+        .into();
+
+        ConnectorManifest {
+            capabilities: ConnectorCapabilities {
+                read: HashSet::from([
+                    ReadCapabilities::Assets,
+                    ReadCapabilities::Groups,
+                    ReadCapabilities::Policies {
+                        default_policies: true,
+                    },
+                    ReadCapabilities::Users,
+                ]),
+                write: HashSet::from([
+                    WriteCapabilities::Groups { nested: false },
+                    WriteCapabilities::Policies {
+                        default_policies: true,
+                    },
+                ]),
+            },
+            asset_privileges,
+        }
+    }
+
+    fn plan_changes(&self, diffs: &LocalConnectorDiffs) -> Vec<String> {
+        match self.generate_request_plan(diffs) {
+            Ok(plan) => plan.flatten_to_string_vec(),
+            Err(err) => {
+                error!("Unable to generate plan for Tableau: {err}");
+                vec![]
+            }
+        }
+    }
+
+    async fn apply_changes(&self, diffs: &LocalConnectorDiffs) -> Result<String> {
+        let mut success_counter = 0;
+        let mut failure_counter = 0;
+        // This is designed in such a way that each query_set may be run concurrently.
+        let futures = self.generate_plan_futures(diffs)?;
+
+        for request_set in [futures.0, futures.1, futures.2] {
+            let results = futures::stream::iter(request_set)
+                .buffered(coordinator::CONCURRENT_METADATA_FETCHES)
+                .collect::<Vec<_>>()
+                .await;
+
+            for result in results {
+                match result {
+                    Err(e) => {
+                        error!("{:?}", e);
+                        failure_counter += 1;
+                    }
+                    Ok(_) => {
+                        success_counter += 1;
+                    }
+                }
+            }
+        }
+        Ok(format!(
+            "{} successful requests\n{} failed requests",
+            success_counter, failure_counter
+        ))
     }
 }
